/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: compile.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
|*===================================================================*|
|*
|*	$Log: compile.c,v $
|*	Revision 1.1  2018/12/21 00:14:50  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.6  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.5  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.4  2008/12/12 22:30:58  wolf
|*	*** empty log message ***
|*
|*	Revision 1.3  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/

#include "nano.h"

#include "compile.h"

#include "tree.h"
#include "code.h"


/****************************************************************\
|*								*|
|*	Hilfsroutinen zur Codeerzeugung: Einzelbytes		*|
|*								*|
|*--------------------------------------------------------------*|
|*								*|
|*	Die interne Adressierung "ireg" steht für folgendes:	*|
|*								*|
|*	 00- 7f:  interner Speicher				*|
|*	 80- ff:  interne Register				*|
|*								*|
|*	    100:  Durch DPTR adressierter Programm-Speicher	*|
|*	    101:  Durch DPTR adressierter externer Speicher	*|
|*	    102:  Durch P2:R0 adressierter externer Speicher	*|
|*	    103:  Durch P2:R1 adressierter externer Speicher	*|
|*								*|
|*	    104:  Konstante (aus zusätzlichem Argument)		*|
|*	    106:  interner Speicher adressiert durch R0		*|
|*	    107:  interner Speicher adressiert durch R1		*|
|*	108-10f:  Register R0 bis R7				*|
|*								*|
|*--------------------------------------------------------------*|
|*	100-103 sind begrenzt auf "Load/Store-Acc" und "IncAdr"	*|
\****************************************************************/


/*------------------------------------------------------*\
|*	Führe eine Operation mit dem Akku aus.		*|
|*	Der Prefix "x0" in opcode wird aufgefüllt mit	*|
|*		der Adressierungsart.			*|
\*------------------------------------------------------*/

static BOOL
CodeOpOk (unsigned ireg) {
	return (ireg >= 0x106 && ireg <= 0x10f) || ireg <= 0xff;
}


static void
CodeOp (uint8 code, unsigned arg, uint8 data) {

	/*----------------------------------------------*\
	|*	Konstante: Adressierungscode x4		*|
	\*----------------------------------------------*/

	if (arg==0x104) {
		if (code==0x40 && data==0) return;
		if (code==0x60 && data==0) return;
		Code2 (code | 0x04, data);
		return;
	}

	/*----------------------------------------------*\
	|*	Direkter Speicher: Adressierungscode x5	*|
	\*----------------------------------------------*/

	if (arg<=0xff) {
		Code2 (code | 0x05, arg);
		return;
	}

	/*----------------------------------------------*\
	|*	@Ri, Ri: Adressierungscodes x6..xF	*|
	\*----------------------------------------------*/

	if (arg>=0x106 && arg<=0x10f) {
		Code1 (code | (arg&0x0f));
		return;
	}

	/*----------------------------------------------*\
	|*	Geht nicht				*|
	\*----------------------------------------------*/

	Error (MsgInternal, "CodeOp");
}


/*------------------------------------------------------*\
|*	Lade den Akku von der angegebenen Adresse	*|
\*------------------------------------------------------*/

static void
CodeLoadAcc (unsigned ireg, uint8 data) {

	switch (ireg) {

	case ACC:
		return;
	case B:
		Code2 (0xc5, B);
		return;
	case DPC:
		Code1 (0xe4);
		Code1 (0x93);
		return;
	case DPD:
		Code1 (0xe0);
		return;
	case R0D:
		Code1 (0xe2);
		return;
	case R1D:
		Code1 (0xe3);
		return;
	case 0x104:
		if (data) Code2 (0x74, data);
		    else  Code1 (0xe4);
		return;
	default:
		if (ireg<=0xff) {
			Code2 (0xe5, ireg);
			return;
		}

		if (ireg>=0x106 && ireg<=0x10f) {
			Code1 ((ireg&0x0f) | 0xe0);
			return;
		}
	}

	Error (MsgInternal, "CodeLoadAcc");
}


/*------------------------------------------------------*\
|*	Speichere den Akku auf die angegebenen Adresse	*|
\*------------------------------------------------------*/

static void
CodeStoreAcc (unsigned ireg) {

	switch (ireg) {

	case ACC:
		return;
	case B:
		Code2 (0xc5, B);
		return;
	case DPC:
		break;
	case DPD:
		Code1 (0xf0);
		return;
	case R0D:
		Code1 (0xf2);
		return;
	case R1D:
		Code1 (0xf3);
		return;
	default:
		if (ireg<=0xff) {
			Code2 (0xf5, ireg);
			return;
		}

		if (ireg>=0x106 && ireg<=0x10f) {
			Code1 ((ireg&0x0f) | 0xf0);
			return;
		}
	}
	Error (MsgInternal, "CodeStoreAcc");
}


/*------------------------------------------------------*\
|*	Bewege ein Byte, wenn noetig ueber den Akku	*|
\*------------------------------------------------------*/

static void
CodeMoveByte (unsigned dest, unsigned src) {

	if (dest==src) {
		if (dest<MD0||dest>MD5) return;
		return;
	}

	if (dest==ACC) {
		CodeLoadAcc (src, 0);
		return;
	}

	if (src==ACC) {
		CodeStoreAcc (dest);
		return;
	}

	if (dest<=255 && src<=255) {
		Code3 (0x85, src, dest);
		return;
	}

	if (dest<=255 && src>=0x106 && src<=0x10f) {
		Code2 ((src&0xf) | 0x80, dest);
		return;
	}

	if (src<=255 && dest>=0x106 && dest<=0x10f) {
		Code2 ((dest&0xf) | 0xa0, src);
		return;
	}

	CodeLoadAcc (src, 0);	 /* lade in Acc */
	CodeStoreAcc(dest);	 /* speichere Acc nach dest */
}


/*------------------------------------------------------*\
|*	Lade eine ein-Byte-Konstante			*|
\*------------------------------------------------------*/

static void
CodeMoveConstant (unsigned ireg, uint8 data) {

	if (ireg==ACC) {
		if (data) Code2 (0x74, data); else Code1 (0xe4);
		return;
	}

	if (ireg<=0xff) {
		Code3 (0x75, ireg, data);
		return;
	}

	if (ireg>=0x106 && ireg<=0x10f) {
		Code2 ((ireg&0xf) | 0x70, data);
		return;
	}

	if (data) Code2 (0x74, data); else Code1 (0xe4);
	CodeStoreAcc (ireg);
}


/*------------------------------------------------------*\
|*	Bestimme die nachfolgende interne Adresse zu	*|
|*	einer gegebenen Adresse.			*|
|*	Erzeuge bei Bedarf den Code zur Aenderung	*|
\*------------------------------------------------------*/

static void
IncAdr (unsigned* aptr) {

	switch (*aptr) {

	case ACC:
		*aptr = B;
		return;
	case DPC:
	case DPD:
		Code1 (0xa3);
		return;
	case R0D:
	case R0L:
		Code1 (0x8);
		return;
	case R1D:
	case R1L:
		Code1 (0x9);
		return;
	default:
		(*aptr)++;
	case 0x104:
		return;
	}
	Error (MsgInternal, "IncAdr");
}

/*------------------------------------------------------*\
|*	Bestimme die vorhergehende interne Adresse zu	*|
|*	einer gegebenen Adresse.			*|
|*	Erzeuge bei Bedarf den Code zur Aenderung	*|
\*------------------------------------------------------*/

static void
DecAdr (unsigned* aptr) {

	switch (*aptr) {
	case B:
		*aptr = ACC;
		return;
	case DPC:
	case DPD:
		Code2 (0x05, 0x82);
		Code3 (0xd5, 0x82, 0x02);
		Code2 (0x15, 0x83);
		Code2 (0x15, 0x82);
		return;
	case R0D:
	case R0L:
		Code1 (0x18);
		return;
	case R1D:
	case R1L:
		Code1 (0x19);
		return;
	default:
		(*aptr)--;
	case 0x104:
		return;
	}
	Error (MsgInternal, "DecAdr");
}


/*------------------------------------------------------*\
|*	Bewege ein Byte oder ein Wort,			*|
|*	Wenn noetig, mit Groessenerweiterung.		*|
\*------------------------------------------------------*/

static void
CodeMove (unsigned dest, unsigned src, unsigned destsize, unsigned srcsize) {

	if ((src==DPC || src==DPD) && (dest==DPL) && destsize>=2 && srcsize>=2) {

		/*------------------------------------------------------*\
		|*	DPTR = DPTR^ benötigt einen Zwischenspeicher	*|
		\*------------------------------------------------------*/
		CodeMoveByte(B, src);
		IncAdr(&src);
		IncAdr(&dest);
		CodeMoveByte(dest, src);
		DecAdr(&dest);
		CodeMoveByte(dest, B);
		return;
	}

	/*------------------------------------------------------*\
	|*	Kopiere erstes Byte				*|
	\*------------------------------------------------------*/

	if (dest != src) CodeMoveByte (dest, src);

	if (destsize<=1) return;

	/*------------------------------------------------------*\
	|*	Zweite Byte gewünscht, aber nichts da		*|
	\*------------------------------------------------------*/

	if (srcsize<=1) {
		IncAdr(&dest);
		CodeMoveConstant(dest, 0);
		return;
	}

	/*------------------------------------------------------*\
	|*	Zweite Byte gewünscht und auch vorhanden	*|
	\*------------------------------------------------------*/

	if (dest==src) return;
	IncAdr(&dest);
	IncAdr(&src);
	if (dest==B && (src==DPD || src==DPC)) Code2 (0xc5, B);
	CodeMoveByte (dest, src);
}


/****************************************************************\
|****************************************************************|
|***								*|
|***								*|
|***		Vorbereitungsroutinen				*|
|***								*|
|***								*|
|****************************************************************|
\****************************************************************/


static int
ComposeBitAddress (unsigned iaddr, unsigned bitnum) {

	if (iaddr>0xff || bitnum>7)
		return -1;

	if ((iaddr & 0xf0) == 0x20)
		return (iaddr<<3) | bitnum;

	if ((iaddr & 0x87) == 0x80)
		return iaddr | bitnum;

	return -1;
}


static void
CompileNode (Node node, unsigned size, unsigned dest, unsigned freereg);


/*==============================================================*\
|*								*|
|*	Liefere die interne Adresse zum Zugriff auf ein Objekt	*|
|*	Erzeugt den Code fuer die Vorbereitung (Rekursion)	*|
|*								*|
\*==============================================================*/

static unsigned
CompileOperand (Node node, unsigned freereg, unsigned defaultdest) {

#if 0
	fprintf (errfile, "CompileOperand (node=%p free=%x defaultdest=%x)\n",
		(void*) node, freereg, defaultdest);
#endif

	switch (node->op) {
	case cval:
		return 0x104;

	case mread:
		switch ((Storage) node->value) {

		case reg:
			if (node->left->op!=cval) Error (MsgInternal, "CompileRightOperand");
			return node->left->value;
			break;

		case internal:
			if (node->left->op==cval && node->left->value <= 0x7f)
				return node->left->value;
			CompileNode (node, 1, R1, freereg);
			return R1L;
		default:
			CompileNode (node, node->isize, defaultdest, freereg);
			return defaultdest;
		}
	default: ;
	}

	CompileNode (node, node->isize, defaultdest, freereg);
	return defaultdest;
}


/*==============================================================*\
|*								*|
|*	Liefere die interne Adresse zum Zugriff auf das		*|
|*	rechte Argument.					*|
|*	Erzeugt den Code fuer die Vorbereitung (Rekursion)	*|
|*								*|
\*==============================================================*/

static unsigned
CompileRightOperand (Node node, unsigned freereg, unsigned defaultdest) {

	Node right = node->right;

	switch (right->op) {
	case cval:
		return 0x104;

	case mread:
		switch ((Storage) right->value) {

		case reg:
			if (right->left->op!=cval) Error (MsgInternal, "CompileRightOperand");
			return right->left->value;
			break;

		case internal:
			if (right->left->op==cval && right->left->value <= 0x7f)
				return right->left->value;
			if (!node->left || !(node->left->sregs & sreg_ri)) {
				CompileNode (right->left, 1, R1, freereg);
				return R1L;
			}
			/* fall thru */
		default:
			CompileNode (right, node->isize, defaultdest, freereg);
			return defaultdest;
		}
	default: ;
	}

	CompileNode (right, node->isize, defaultdest, freereg);
	return defaultdest;
}


/*==============================================================*\
|*								*|
|*	Liefere die interne Adresse zum Zugriff auf das		*|
|*	linke Argument.					*|
|*	Erzeugt den Code fuer die Vorbereitung (Rekursion)	*|
|*								*|
\*==============================================================*/


static unsigned
CompileLeftOperand (Node node, unsigned freereg, unsigned rightarg) {

	Node left  = node->left;

	switch (left->op) {
	case cval:
		return 0x104;

	case mread:
		switch ((Storage) left->value) {

		case reg:
			if (left->left->op!=cval) Error (MsgInternal, "CompileLeftOperand");
			return left->left->value;

		case internal:
			if (left->left->op==cval && left->left->value <= 0x7f)
				return left->left->value;
			CompileNode (left->left, 1, R1, freereg);
			return R1L;

		case external:
			if (left->left->op!=cval) break;
			CompileNode (left->left, 2, DPL, freereg);
			return DPD;

		default: ;
		}
	default: ;
	}

	CompileNode (left, node->isize, ACC,
		rightarg>=freereg ? freereg+node->right->osize : freereg);
	return ACC;
}


static BOOL
CompileMemoryReference (Node node, unsigned*addr) {

	Node left = node->left;

	switch ((Storage) node->value) {

	case reg:
		if (left->op!=cval) break;
		*addr = left->value;
		return TRUE;
	
	case internal:
		if (left->op==cval && left->value<=0x7f) {
			*addr = left->value;
			return TRUE;
		}
		CompileNode(left, 1, R0, R0);
		*addr = R0L;
		return TRUE;

	case external:
		CompileNode(left, 2, DPL, R0);
		*addr = DPD;
		return TRUE;

	default: ;
	}
	return FALSE;
}


/****************************************************************\
|****************************************************************|
|***								*|
|***								*|
|***		Uebersetzung bestimmter Knotentype		*|
|***								*|
|***								*|
|****************************************************************|
\****************************************************************/


/*==============================================================*\
|*								*|
|*		Vergleich auf kleiner				*|
|*								*|
\*==============================================================*/

static void
CompileTestLess (Node node, unsigned freereg) {

	unsigned iregr = CompileRightOperand (node, freereg, freereg);
	unsigned iregl = CompileLeftOperand  (node, freereg, iregr);

	Node right = node->right;
	Node left  = node->left;

	/*======================================================*\
	|*	Erzeuge Code fuer das Low Byte			*|
	\*======================================================*/

	Code1 (0xc3);

	CodeLoadAcc (iregl , lo(left->value));
	CodeOp (0x90, iregr, lo(right->value));

	if (node->isize<=1) return;

	/*======================================================*\
	|*	Erzeuge Code fuer das High Byte			*|
	\*======================================================*/

	if (left->osize>=2) {
		IncAdr (&iregl);
		CodeLoadAcc (iregl, hi(left->value));
	} else {
		Code1 (0xe4);
	}

	if (right->osize>=2) {
		IncAdr (&iregr);
		CodeOp (0x90, iregr, hi(right->value));
	} else {
		CodeOp (0x90, 0x104, 0);
	}
}


/*==============================================================*\
|*								*|
|*		Vergleich auf Gleichheit			*|
|*								*|
\*==============================================================*/

static void
CompileTestEqual (Node node, unsigned freereg) {

	unsigned iregr = CompileRightOperand (node, freereg, freereg);
	unsigned iregl = CompileLeftOperand  (node, freereg, iregr);

	Node right = node->right;
	Node left  = node->left;

	/*======================================================*\
	|*	Erzeuge Code fuer das Low Byte			*|
	\*======================================================*/

	CodeLoadAcc (iregl , lo(left->value));
	CodeOp (0x60, iregr, lo(right->value));

	if (node->isize<=1) return;

	/*======================================================*\
	|*	Erzeuge Code fuer das High Byte			*|
	\*======================================================*/

	Label skip;
	LabelInit (&skip);

	CodeJump (&skip, 0x70, 0);

	if (left->osize>=2) {
		IncAdr (&iregl);
		CodeLoadAcc (iregl, hi(left->value));

		if (right->osize>=2) {
			IncAdr (&iregr);
			CodeOp (0x60, iregr, hi(right->value));
		}
	} else {
		IncAdr (&iregr);
		CodeLoadAcc (iregr, hi(right->value));
	}

	CodeLabel (&skip);
}


/*==============================================================*\
|*								*|
|*		Lineare Operatoren (+,-,|,&,^)			*|
|*								*|
\*==============================================================*/

static void
CompileOperator (Node node, unsigned size, unsigned dest, unsigned freereg,
	uint8 prep, uint8 op1, uint8 op2) {

	unsigned iregr = CompileRightOperand (node, freereg, freereg);
	unsigned iregl = CompileLeftOperand  (node, freereg, iregr);

	Node right = node->right;
	Node left  = node->left;

	/*======================================================*\
	|*	Erzeuge Code fuer das Low Byte			*|
	\*======================================================*/

#if 0
	fprintf (errfile, "CompileOperator: size=%u isize=%u osize=%o dest=%x\n",
		size, node->isize, node->osize, dest);
#endif

	if (prep) Code1 (prep);

	CodeLoadAcc (iregl , lo(left->value));
	CodeOp (op1, iregr, lo(right->value));
	CodeStoreAcc (dest);

	if (size <= 1) return;

	/*======================================================*\
	|*	Fülle das High Byte mit einer 0			*|
	\*======================================================*/

	if (node->osize <= 1) {
		IncAdr (&dest);
		CodeMoveConstant (dest, 0);
		return;
	}

	/*======================================================*\
	|*	Erzeuge Code fuer das High Byte			*|
	\*======================================================*/

#if 0
	fprintf (errfile, "CompileOperator byte2 dest=%x iregl=%x iregr=%x\n",
			dest, iregl, iregr);
#endif

	if (left->osize>=2) {
		if (dest==ACC && iregl != ACC) Code2 (0xc5, B);
		IncAdr (&iregl);
		CodeLoadAcc (iregl, hi(left->value));
	} else {
		if (dest==ACC) Code2 (0xc5, B);
		if ((op2==0x40 || op2==0x60) && right->osize>=2) {
			op2 = 0xe0;
		} else {
			CodeLoadAcc (0x104, 0);
		}
	}

	if (right->osize>=2) {
		IncAdr (&iregr);
		CodeOp (op2, iregr, hi(right->value));
	} else {
		CodeOp (op2, 0x104, hi(right->value));
	}

	IncAdr (&dest);
	CodeStoreAcc (dest);
}


/*==============================================================*\
|*								*|
|*		Multiplikation und Division			*|
|*								*|
\*==============================================================*/

static void
CompileMulDivMod (Node node, unsigned dest, unsigned freereg,
	uint8 op, unsigned resreg, unsigned resize) {

#if 0
	fprintf (errfile, "CompileMulDivMod osize=%u isize=%u resize=%u\n",
		node->osize, node->isize, resize);
#endif

	unsigned iregr = CompileRightOperand (node, freereg, freereg);
	unsigned iregl = CompileLeftOperand  (node, freereg, iregr);

	Node right = node->right;
	Node left  = node->left;

	/*======================================================*\
	|*	Erzeuge Code					*|
	\*======================================================*/

	CodeLoadAcc (iregl, lo(left->value));
	if (iregr==0x104)
		Code3 (0x75, B, lo(right->value));
	else
		CodeMove (B, iregr, 1, 1);
	Code1 (op);
	CodeMove (dest, resreg, node->osize, resize);
}


/****************************************************************\
|****************************************************************|
|***								*|
|***								*|
|***		Uebersetzung eines Baumknotens			*|
|***								*|
|***								*|
|****************************************************************|
\****************************************************************/


static void
CompileNode (Node node, unsigned size, unsigned dest, unsigned freereg) {

#if 0
 	fprintf (errfile, "CompileNode node=%p size=%u dest=%x freereg=%x\n",
		(void*)node, size, dest, freereg);
#endif

	Node left = node->left;
	Node right= node->right;

	unsigned addr=0;

	switch (node->op) {

	/*------------------------------------------------------*\
	|*	Konstanter Wert					*|
	\*------------------------------------------------------*/

	case cval:
		if (dest==DPL && size==2) {
			Code3 (0x90, hi(node->value), lo(node->value));
			return;
		}

		CodeMoveConstant (dest, lo(node->value));
		if (size<2) return;
		IncAdr (&dest);
		CodeMoveConstant (dest, hi(node->value));
		return;

	/*------------------------------------------------------*\
	|*	Variable lesen					*|
	\*------------------------------------------------------*/

	case mread:
		switch ((Storage) node->value) {

		case internal:
			if (left->op==cval && left->value<0x7f) {
				CodeMove(dest, left->value, size, node->osize);
			} else {
				CompileNode(left, 1, R1, freereg);
				CodeMove(dest, R1L, size, node->osize);
			}
			return;

		case reg:
			if (left->op != cval) break;
			CodeMove (dest, left->value, size, node->osize);
			return;

		case code:
			CompileNode (left, 2, DPL, freereg);
			CodeMove (dest, DPC, size, node->osize);
			return;

		case external:
			if (left->op==cval) {
				Code3 (0x90, hi(left->value), lo(left->value));
			} else {
				CompileNode (left, 2, DPL, freereg);
			}
			CodeMove (dest, DPD, size, node->osize);
			return;

		default: ;
		}
		break;

	/*------------------------------------------------------*\
	|*	Addition und Subtraktion			*|
	\*------------------------------------------------------*/

	case addu:
		if (size==1 && right->op==cval) {
			switch (lo(right->value)) {
			case 0:
                		CompileNode (node->left, 1, dest, freereg);
				return;
			case 1:
                		CompileNode (node->left, 1, ACC, freereg);
                		Code1 (0x04);
                		CodeStoreAcc (dest);
				return;
			case 0xff:
                		CompileNode (node->left, 1, ACC, freereg);
                		Code1 (0x14);
                		CodeStoreAcc (dest);
				return;
			}
		}

		CompileOperator(node, size, dest, freereg, 0x00, 0x20, 0x30);
		return;

		break;

	case adduc:
		CompileOperator(node, size, dest, freereg, 0x00, 0x30, 0x30);
		return;

	case subu:
		if (size==1 && right->op==cval) {
			switch (lo(right->value)) {
			case 0:
                		CompileNode (node->left, 1, dest, freereg);
				return;
			case 1:
                		CompileNode (node->left, 1, ACC, freereg);
                		Code1 (0x14);
                		CodeStoreAcc (dest);
				return;
			case 0xff:
                		CompileNode (node->left, 1, ACC, freereg);
                		Code1 (0x04);
                		CodeStoreAcc (dest);
				return;
			}
		}
		if (size==1 && left->op==cval) {
			switch (lo(left->value)) {
			case 0:
                		CompileNode (node->right, 1, ACC, freereg);
                		Code1 (0xf4);
                		Code1 (0x04);
                		CodeStoreAcc (dest);
				return;
			case 0xff:
                		CompileNode (node->right, 1, ACC, freereg);
                		Code1 (0xf4);
                		CodeStoreAcc (dest);
				return;
			}
		}

		CompileOperator(node, size, dest, freereg, 0xc3, 0x90, 0x90);
		return;

	case subub:
		CompileOperator(node, size, dest, freereg, 0x00, 0x90, 0x90);
		return;

	/*------------------------------------------------------*\
	|*	Bitweise logische Operationen			*|
	\*------------------------------------------------------*/

	case oru:
		if (size==1 && right->op==cval) {
			switch (lo(right->value)) {
			case 0:
                		CompileNode (node->left, 1, dest, freereg);
				return;
			case 0xff:
				CodeMoveConstant (dest, 0xff);
				return;
			}
		}
		CompileOperator(node, size, dest, freereg, 0x00, 0x40, 0x40);
		return;

	case andu:
		if (size==1 && right->op==cval) {
			switch (lo(right->value)) {
			case 0:
				CodeMoveConstant (dest, 0xff);
				return;
			case 0xff:
                		CompileNode (node->left, 1, dest, freereg);
				return;
			}
		}
		CompileOperator(node, size, dest, freereg, 0x00, 0x50, 0x50);
		return;

	case xoru:
		if (size==1 && right->op==cval) {
			switch (lo(right->value)) {
			case 0:
                		CompileNode (node->left, 1, dest, freereg);
				return;
			case 0xff:
                		CompileNode (node->left, 1, ACC, freereg);
                		Code1 (0xf4);
                		CodeStoreAcc (dest);
				return;
			}
		}
		CompileOperator(node, size, dest, freereg, 0x00, 0x60, 0x60);
		return;

	/*------------------------------------------------------*\
	|*	Mul Div Mod					*|
	\*------------------------------------------------------*/

	case mulu:
		CompileMulDivMod (node, dest, freereg, 0xa4, ACC, 2);
		return;

	case divu:
		CompileMulDivMod (node, dest, freereg, 0x84, ACC, 1);
		return;

	case modu:
		CompileMulDivMod (node, dest, freereg, 0x84, B, 1);
		return;

	/*------------------------------------------------------*\
	|*	Schiebe links					*|
	\*------------------------------------------------------*/

	case shl1u:
		CompileNode (right, node->isize, ACC, freereg);
		Code1 (0xc3);	/* CLR C	*/
		Code1 (0x33);	/* RLC A	*/
		CodeStoreAcc (dest);

		if (node->osize <= 1) return;
		
		Code2 (0xc5, B);
		if (node->isize <= 1) Code1 (0xe4);

		Code1 (0x33);	/* RLC A	*/
		IncAdr (&dest);
		CodeStoreAcc (dest);
		return;

	/*------------------------------------------------------*\
	|*	Bitweise komplement				*|
	\*------------------------------------------------------*/

	case cplu:
		addr = CompileRightOperand (node, freereg, ACC);
		CodeLoadAcc (addr, 0);
		Code1 (0xf4);
		CodeStoreAcc (dest);

		if (size<=1) return;

		IncAdr (&dest);

		if (node->isize<=1) {
			CodeMoveConstant (dest, 0xff);
			return;
		}

		IncAdr (&addr);
		CodeLoadAcc (addr, 0);
		Code1 (0xf4);
		CodeStoreAcc (dest);
		return;

	/*------------------------------------------------------*\
	|*	Logisches Nicht					*|
	\*------------------------------------------------------*/

	case notu:
		addr = CompileRightOperand (node, freereg, ACC);
		CompileNode(node->right, node->isize, ACC, freereg);
		CodeLoadAcc (addr, lo(right->value));
		if (right->osize>=2) {
			IncAdr (&addr);
			CodeOp (0x40, addr, hi(right->value));
		}
		Code2 (0x60,    2);
		Code2 (0x74, 0xff);
		if (CodeTrue==0xff) Code1 (0xf4); else Code1 (0x14);
		CodeMove(dest, ACC, node->osize, 1);
		return;

	/*------------------------------------------------------*\
	|*	Vergleiche	==	!=			*|
	\*------------------------------------------------------*/

	case eq:
		CompileTestEqual (node, freereg);
		Code2 (0x60,    2);
		Code2 (0x74, 0xff);
		if (CodeTrue==0xff) Code1 (0xf4); else Code1 (0x14);
		CodeMove(dest, ACC, node->osize, 1);
		return;

	case ne:
		CompileTestEqual (node, freereg);
		if (CodeTrue==0xff) {
			Code2 (0x60,    2);
			Code2 (0x74, 0xff);
		} else {
			Code2 (0x60,    2);
			Code2 (0x74, 0x01);
		}
		CodeMove(dest, ACC, node->osize, 1);
		return;

	/*------------------------------------------------------*\
	|*	Vergleiche	<	>=			*|
	\*------------------------------------------------------*/

	case ltu:
		CompileTestLess (node, freereg);
		if (CodeTrue==0xff) {
			Code1 (0xe4);		/* CLR A	*/
			Code2 (0x94, 9);	/* SUBB A,0	*/
		} else {
			Code1 (0xe4);		/* CLR A	*/
			Code1 (0x33);		/* RLC A	*/
		}
		CodeMove(dest, ACC, node->osize, 1);
		return;

	case geu:
		CompileTestLess (node, freereg);
		Code1 (0xb3);			/* CPL C	*/
		if (CodeTrue==0xff) {
			Code1 (0xe4);		/* CLR A	*/
			Code2 (0x94, 9);	/* SUBB A,0	*/
		} else {
			Code1 (0xe4);		/* CLR A	*/
			Code1 (0x33);		/* RLC A	*/
		}
		CodeMove(dest, ACC, node->osize, 1);
		return;


	default: ;
	}
	char buffer[30];
	snprintf(buffer, 30, "CompileNode(%u)", node->op);
	Error (MsgInternal, buffer);
}



/****************************************************************\
|****************************************************************|
|***								*|
|***								*|
|***		Uebersetzung eines Ausdruckbaumes		*|
|***								*|
|***								*|
|****************************************************************|
\****************************************************************/


void
CompileTree (void) {

	/*------------------------------------------------------*\
	|*	Wurzelknoten des Baumes				*|
	\*------------------------------------------------------*/

	Node node = getLastUnused();

	/*------------------------------------------------------*\
	|*	Groessenreduktion der Knoten			*|
	|*------------------------------------------------------*|
	|*	Bei 8031-Prozessoren muessen 16-bit-Operationen	*|
	|*	aus mehreren 8-bit-Operationen zusammengesetzt	*|
	|*	werden. Das ist Zeit- und Speicher-aufwendig.	*|
	|*	Deshalb wird 16-bit zu 8-bit reduziert,		*|
	|*	wenn die oberen 8bit nicht gebraucht werden.	*|
	\*------------------------------------------------------*/

	OptimizeNodes ();

	/*------------------------------------------------------*\
	|*	Compile						*|
	\*------------------------------------------------------*/

	Node left = node->left;
	Node right= node->right;

	unsigned addr=0;
	int bitaddr=-1;

	switch (node->op) {

	/*------------------------------------------------------*\
	|*	Speicher					*|
	\*------------------------------------------------------*/

	case mwrite:
		switch ((Storage) node->value) {

		case internal:
			if (left->op==cval && left->value<0x80) {
				CompileNode(right, node->osize, left->value, R2);
				return;
			}
			CompileNode(left, 1, R0, R2);
			CompileNode(right, node->osize, R0L, R2);
			return;

		case reg:
			if (left->op != cval) break;
			CompileNode (right, node->osize, left->value, R2);
			return;

		case external:
			if (!(right->sregs & sreg_dptr)) {
				CompileNode (left, node->isize, DPL, R2);
				CompileNode (right,node->osize, DPD, R2);
				return;
			}
			if (left->nregs<0 && node->osize<=1) {
				CompileNode (right,right->osize, ACC, R2);
				CompileNode (left, node->isize, DPL, R2);
				CodeMove(DPD, ACC, node->osize, right->osize);
				return;
			}
			CompileNode (right,right->osize, R2, R2);
			CompileNode (left, node->isize, DPL, R2+right->osize);
			CodeMove(DPD, R2, node->osize, right->osize);
			return;

		default: ;
		}
		break;

	/*------------------------------------------------------*\
	|*	Inkrementiere Variable im Speicher		*|
	\*------------------------------------------------------*/

	case mincr1u:

		if (!CompileMemoryReference(node, &addr)) break;

		/*------------------------------------------------------*\
		|*	Spezialfall: inkrementiere Datenpointer		*|
		\*------------------------------------------------------*/

		if (addr==DPL && node->osize==2) {
			Code1 (0xa3);
			return;
		}

		/*------------------------------------------------------*\
		|*	Spezialfall: inkrementiere Akku			*|
		\*------------------------------------------------------*/

		if (addr==ACC && node->osize==1) {
			Code1 (0x04);
			return;
		}

		/*------------------------------------------------------*\
		|*	Interner Speicher, @Ri oder Ri			*|
		\*------------------------------------------------------*/

		if (CodeOpOk (addr)) {

			/*----------------------------------------------*\
			|*	Niederwertiges Byte			*|
			\*----------------------------------------------*/

			CodeOp (0x00, addr, 0);

			if (node->osize <= 1) return;

			/*----------------------------------------------*\
			|*	Höherwertiges Byte			*|
			\*----------------------------------------------*/

			Label l;
			LabelInit (&l);
			CodeLoadAcc (addr, 0);
			CodeJump (&l, 0x70, 0);
			IncAdr(&addr);
			CodeOp (0x00, addr, 0);
			CodeLabel (&l);
			return;
		}

		/*------------------------------------------------------*\
		|*	Sonstige Adressierungsarten			*|
		\*------------------------------------------------------*/

		/*----------------------------------------------*\
		|*	Niederwertiges Byte			*|
		\*----------------------------------------------*/

		CodeLoadAcc (addr, 0);
		Code1 (0x04);
		CodeStoreAcc (addr);

		if (node->osize >= 1) {

			/*--------------------------------------*\
			|*	Höherwertiges Byte		*|
			\*--------------------------------------*/

			Label l;
			LabelInit (&l);
			CodeJump (&l, 0x70, 0);
			IncAdr(&addr);
			CodeLoadAcc (addr, 0);
			Code1 (0x04);
			CodeStoreAcc (addr);
			CodeLabel (&l);
		}

		return;

	/*------------------------------------------------------*\
	|*	Dekrementiere Variable im Speicher		*|
	\*------------------------------------------------------*/

	case mdecr1u:

		if (!CompileMemoryReference(node, &addr)) break;

		/*------------------------------------------------------*\
		|*	Spezialfall: inkrementiere Akku			*|
		\*------------------------------------------------------*/

		if (addr==ACC && node->osize==1) {
			Code1 (0x14);
			return;
		}

		/*------------------------------------------------------*\
		|*	Interner Speicher, @Ri oder Ri			*|
		\*------------------------------------------------------*/

		if (CodeOpOk (addr)) {

			/*----------------------------------------------*\
			|*	Niederwertiges Byte			*|
			\*----------------------------------------------*/

			CodeOp (0x10, addr, 0);

			if (node->osize <= 1) return;

			/*----------------------------------------------*\
			|*	Höherwertiges Byte			*|
			\*----------------------------------------------*/

			Label l;
			LabelInit (&l);
			CodeLoadAcc (addr, 0);
			CodeJump (&l, 0xb4, 0xff);	/* CJNE A, 0xff */
			IncAdr(&addr);
			CodeOp (0x10, addr, 0);
			CodeLabel (&l);
			return;
		}

		/*------------------------------------------------------*\
		|*	Sonstige Adressierungsarten			*|
		\*------------------------------------------------------*/

		/*----------------------------------------------*\
		|*	Niederwertiges Byte			*|
		\*----------------------------------------------*/

		CodeLoadAcc (addr, 0);
		Code1 (0x14);
		CodeStoreAcc (addr);

		if (node->osize >= 1) {

			/*--------------------------------------*\
			|*	Höherwertiges Byte		*|
			\*--------------------------------------*/

			Label l;
			LabelInit (&l);
			CodeJump (&l, 0xb4, 0xff);	/* CJNE A, 0xff */
			IncAdr(&addr);
			CodeLoadAcc (addr, 0);
			Code1 (0x04);
			CodeStoreAcc (addr);
			CodeLabel (&l);
		}

		return;

	/*------------------------------------------------------*\
	|*	Schiebe-Links im Speicher			*|
	\*------------------------------------------------------*/

	case mshl1u:

		if (!CompileMemoryReference(node, &addr)) break;

		Code1 (0xc3);	 /* CLR C */
		CodeLoadAcc (addr, 0);
		Code1 (0x33);	 /* RLC A */
		CodeStoreAcc(addr);
	
		if (node->osize<=1) return;
	
		IncAdr(&addr);
	
		CodeLoadAcc (addr, 0);
		Code1 (0x33);	 /* RLC A */
		CodeStoreAcc(addr);
		return;

	/*------------------------------------------------------*\
	|*	Schiebe-Rechts im Speicher			*|
	\*------------------------------------------------------*/

	case mshr1u:

		if (!CompileMemoryReference(node, &addr)) break;

		Code1(0xc3);	 /* CLR C */

		if (node->osize>=2) {

			IncAdr(&addr);

			CodeLoadAcc (addr, 0);
			Code1 (0x13);	 /* RRC A */
			CodeStoreAcc(addr);

			DecAdr(&addr);
		}

		CodeLoadAcc (addr, 0);
		Code1(0x13);	 /* RRC A */
		CodeStoreAcc(addr);
		return;

	/*------------------------------------------------------*\
	|*	Bit-Operationen im Speicher			*|
	\*------------------------------------------------------*/

	case mbitset:
	case mbitclr:
	case mbitcpl:

		if (!CompileMemoryReference(node, &addr)) break;
		if (right->op != cval || right->value>7) break;

		/*----------------------------------------------*\
		|*	Bitadressierbares Byte?			*|
		\*----------------------------------------------*/

		bitaddr = ComposeBitAddress(addr, right->value);

		if (bitaddr == 0xd7) {
			switch (node->op) {
			default:
			case mbitset: Code1 (0xd3);
				return;
			case mbitclr: Code1 (0xc3);
				return;
			case mbitcpl: Code1 (0xb3);
				return;
			}
		}

		if (bitaddr >= 0) {
			switch (node->op) {
			default:
			case mbitset: Code2 (0xd2, bitaddr);
				break;
			case mbitclr: Code2 (0xc2, bitaddr);
				break;
			case mbitcpl: Code2 (0xb2, bitaddr);
				break;
			}
			return;
		}

		/*----------------------------------------------*\
		|*	Internes Byte?				*|
		\*----------------------------------------------*/

		if (addr <= 0xff) {

			switch (node->op) {
			default:
			case mbitset: Code3 (0x43, left->value, 1<<right->value);
				break;
			case mbitclr: Code3 (0x53, left->value, ~(1<<right->value));
				break;
			case mbitcpl: Code3 (0x63, left->value, 1<<right->value);
				break;
			}

			return;
		}

		/*----------------------------------------------*\
		|*	Sonstiges Byte				*|
		\*----------------------------------------------*/

		CodeLoadAcc (addr, 0);

		switch (node->op) {
		default:
		case mbitset: Code2 (0xd2, ACC+right->value);
			break;
		case mbitclr: Code2 (0xc2, ACC+right->value);
			break;
		case mbitcpl: Code2 (0xb2, ACC+right->value);
			break;
		}

		CodeStoreAcc (addr);
		return;

	/*------------------------------------------------------*\
	|*	Default: nicht implementiert			*|
	\*------------------------------------------------------*/

	default: ;
	}

	char buffer[30];
	snprintf(buffer, 30, "CompileTree(%u)", node->op);
	Error (MsgInternal, buffer);
}

/*==============================================================*\
|*								*|
|*								*|
|*		Sprung und Schleifenbefehle			*|
|*								*|
|*								*|
\*==============================================================*/


/*----------------------------------------------*\
|*						*|
|*	Uebersetze eine logische Bedingung	*|
|*						*|
\*----------------------------------------------*/

static void
CompileCondJump (Node node, BOOL on, Label * label) {

#if 0
	fprintf (errfile, "CompileCondJump node=%p on=%u label=L%04u laddr=%x\n",
		(void *) node, on, label->lnum, label->laddr);
#endif

	unsigned addr=0;
	int bitaddr = -1;

	Node left  = node->left;
	Node right = node->right;

	Label skip;

	switch (node->op) {

	/*----------------------------------------------*\
	|*	Logische Verknuepfungen			*|
	\*----------------------------------------------*/

	case oru:
		if (right->nregs<0) break;
		if (on) {
			CompileCondJump (left, on, label);
			CompileCondJump (right,on, label);
			return;
		}
		LabelInit (&skip);
		CompileCondJump (left, !on, &skip);
		CompileCondJump (right,on, label);
		CodeLabel (&skip);
		return;

	case andu:
		if (right->nregs<0) break;
		if (!on) {
			CompileCondJump (left, on, label);
			CompileCondJump (right,on, label);
			return;
		}
		LabelInit (&skip);
		CompileCondJump (left, !on, &skip);
		CompileCondJump (right,on, label);
		CodeLabel (&skip);
		return;

	case notu:
		CompileCondJump (right, !on, label);
		return;

	/*----------------------------------------------*\
	|*	Vergleiche				*|
	\*----------------------------------------------*/

	case eq:
		CompileTestEqual (node, R2);
		CodeCondJump (label, on, 0x60, 0x70, 0);
		return;

	case ne:
		CompileTestEqual (node, R2);
		CodeCondJump (label, on, 0x70, 0x60, 0);
		return;

	case ltu:
		CompileTestLess (node, R2);
		CodeCondJump (label, on, 0x40, 0x50, 0);
		return;

	case geu:
		CompileTestLess (node, R2);
		CodeCondJump (label, on, 0x50, 0x40, 0);
		return;

	/*----------------------------------------------*\
	|*	Bit-Tests				*|
	\*----------------------------------------------*/

	case bitis1:

		if (right->op != cval || right->value>7) break;

		/*----------------------------------------------*\
		|*	Möglicherweise direkter Bitzgriff	*|
		\*----------------------------------------------*/

		if (left->op==mread && CompileMemoryReference(left, &addr)) {

			/*----------------------------------------------*\
			  |*	Bitadressierbares Byte?			*|
			  \*----------------------------------------------*/

			bitaddr = ComposeBitAddress(addr, right->value);

			if (bitaddr == 0xd7) {
				CodeCondJump (label, on, 0x40, 0x50, 0);
				return;
			}

			if (bitaddr >= 0) {
				CodeCondJump (label, on, 0x20, 0x30, bitaddr);
				return;
			}

			/*----------------------------------------------*\
			|*	Nein, lade Byte in den Akku		*|
			\*----------------------------------------------*/

			CodeLoadAcc (addr, 0);
			CodeCondJump (label, on, 0x20, 0x30, ACC+right->value);
			return;
		}
		break;

	default: ;
	}

	addr = CompileOperand (node, ACC, R2);

	CodeLoadAcc (addr, lo(node->value));
	if (node->osize>=2) {
		IncAdr (&addr);
		CodeOp (0x40, addr, hi(node->value));
	}

	CodeCondJump (label, on, 0x70, 0x60, 0);
}


/*------------------------------------------------------*\
|*							*|
|*	Uebersetze einen bedingten Sprung 		*|
|*							*|
\*------------------------------------------------------*/

void
CompileCondition (Label * label, BOOL on) {

	Node node =getLastUnused();
	OptimizeNodes ();

	if (getOption('J')==1 || getCodeAddress() - label->laddr < 96) {
		CompileCondJump (node, on, label);
		return;
	}
	Label skip;
	LabelInit (&skip);
	CompileCondJump (node, !on, &skip);
	CodeJump (label, 0, 0);
	CodeLabel (&skip);
}


/*----------------------------------------------------------------------*\
|*									*|
|*	Uebersetze ein "Dekrementier und Springe wenn ungleich Null	*|
|*									*|
\*----------------------------------------------------------------------*/

void
CompileCondNonZero (Label * label) {

	unsigned addr=0;

	Node node =getLastUnused();
	OptimizeNodes ();

	if (node->op==mdecr1u && node->osize<=1 && CompileMemoryReference(node, &addr)) {

		if (addr<=0xff) {

			if (getCodeAddress()+3-label->laddr) {
				CodeJump (label, 0xd5, addr);
				return;
			}
			CodeOp (0x10, addr, 0);
			CodeLoadAcc (addr, 0);

		} else if (addr>=0x108 && addr<=0x10f) {

			if (getCodeAddress()+2-label->laddr) {
				CodeJump (label, 0xd8 | (addr&0x0f), 0);
				return;
			}
			CodeOp (0x10, addr, 0);
			CodeLoadAcc (addr, 0);

		} else {

			CodeLoadAcc (addr, 0);
			Code1 (0x14);
			CodeStoreAcc (addr);
		}

		if (getCodeAddress()+2-label->laddr) {
			CodeJump (label, 0x70, 0);
			return;
		}

		Label skip;
		LabelInit(&skip);
		CodeJump (&skip, 0x60, 0);
		CodeJump (label, 0, 0);
		CodeLabel(&skip);
	}

	Error (MsgInternal, "CompileCondNonZero");
}


/*===================================================================*\
|*	$Id: compile.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
\*===================================================================*/
