/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: nano.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
|*===================================================================*|
|*
|*	$Log: nano.c,v $
|*	Revision 1.1  2018/12/21 00:15:19  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.4  2011/08/09 17:24:50  wolf
|*	+ Usage
|*
|*	Revision 1.3  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*	Revision 1.1  2008/12/07 09:43:42  wolf
|*	Initial revision
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "nano.h"

#include "preprocessor.h"
#include "scanner.h"
#include "declarations.h"
#include "hexfile.h"


/*==============================================================*\
|*								*|
|*	Listing							*|
|*								*|
\*==============================================================*/

FILE * lstfile = NULL;
FILE * errfile = NULL;
int restrictedMode = 0;

void
lstfileselect (const char * group) {
	static const char * active = NULL;
	if (!lstfile || active==group) return;
	if (active) fprintf (lstfile, "\n");
	active = group;
}


/*==============================================================*\
|*								*|
|*	Fehlerbehandlung					*|
|*								*|
\*==============================================================*/

static void
Note (const char note[], const char message[], const char* param) {

	if (lstfile == stdout) lstfileselect (NULL);
	fflush (NULL);
	if (!param) param = "?";
	fprintf (errfile, "%s:%u:%u: %s",
		getCurrentFilename(),
		getCurrentLinenumber(),
		getCurrentColumn(),
		note);
	fprintf (errfile, message, param, strerror(errno));
	fprintf (errfile, "\n\n");
}

static unsigned Errors = 0;

static void
Abort (void) {
	fprintf (errfile, "%s\n", MsgErrorAborted);
	exit (EXIT_FAILURE);
}

void
Fatal (const char message[], const char* param) {
	Note ("Fehler: ", message, param);
	Errors++;
	Abort();
}

void
Error (const char message[], const char* param) {
	Note ("Fehler: ", message, param);
	Errors++;
	if (getOption('e')) Abort();
}

void
Warning (const char message[], const char* param) {
	if (getOption('w')) {
		Error (message, param);
		return;
	}
	Note ("Warnung: ", message, param);
}


/*==============================================================*\
|*								*|
|*	Aufräumen beim Programmende				*|
|*								*|
\*==============================================================*/

const char * hexfilename=0;

void
cleanup (void) {
	if (Errors) {
		fprintf (errfile, MsgError, Errors);
		fprintf (errfile, "\n");

		if (hexfilename && strcmp(hexfilename, "-")) {
			if (!unlink (hexfilename))
				fprintf (errfile, MsgErrorDeleted,
					hexfilename);
				fprintf (errfile, "\n");
		}
	}
}


/*==============================================================*\
|*								*|
|*	Hauptprogramm						*|
|*								*|
\*==============================================================*/

const char Version[] = "$Id: nano.c,v 1.1 2018/12/21 00:15:19 wolf Exp $";

uint32	Options[32] = {0};

static __inline char *
find_ext (const char* base) {
	char * lastslash = strrchr (base, '/');
	char * lastdot = strrchr (lastslash?lastslash+1:base, '.');
	return lastdot ? lastdot : strchr (base, 0);
}

static char usage[] = "Usage:\n\n"
"%s [-V] [-welcst] [-T{cpusecs}] [-L{listfile}] [-H{hexfile}] [-j[1/2]] source.pas\n\n"
"\n"
"\t-V\twrite version and terminate\n"
"\n"
"\t-w\twarning becomes error\n"
"\t-e\tstop on first error\n"
"\n"
"\t-l\twrite listing file \n"
"\t-c\tadd code to listing\n"
"\t-m\tadd messages to listing\n"
"\t-s\tadd symbols to listing\n"
"\t-t\tadd scan tokens to listing\n"
"\n"
"\t-j 1\talways use relative jumps\n"
"\t-j 2\talways use short jumps\n"
"\n";

int
main(int argc, char* argv[]) {

	const char * lstfilename=0;

	errfile = stderr;

	int opt;
	while ((opt = getopt(argc, argv, "H:L:T:welcmstrVj:")) != -1) {
		switch (opt) {
		case 'H':
			hexfilename=optarg;
			break;
		case 'L':
			lstfilename=optarg;
		case 'l':
		case 'w':
		case 'e':
			Options[opt&0x1f]|=1;
			break;
		case 'r':
			restrictedMode = 1;
			break;
		case 'c':
		case 'm':
		case 's':
		case 't':
			Options['l'&0x1f]|=1ul<<(opt&0x1f);
			break;
		case 'j':
			Options[opt&0x1f]=atoi(optarg);
			break;
		case 'V':
			fprintf (stderr, "\n%s\n\n", Version);
			exit (EXIT_SUCCESS);
		case 'T': {
				int limit = atoi(optarg);
				struct rlimit rlimit;
				rlimit.rlim_cur = limit;
				rlimit.rlim_max = limit;
				setrlimit(RLIMIT_CPU, &rlimit);
			}
			break;
		default:
			fprintf (stderr, usage, argv[0]);
			exit (EXIT_FAILURE);
		}
	}

	if (optind != argc-1) {
		fprintf (stderr, usage, argv[0]);
		exit (EXIT_FAILURE);
	}

	/*------------------------------------------------------*\
	|*	Name der Quelldatei				*|
	\*------------------------------------------------------*/

	const char * source_filename = argv[optind];

	/*------------------------------------------------------*\
	|*	Puffer fuer die Namen der Ausgabedateien	*|
	\*------------------------------------------------------*/

	char* filename = (char*) malloc (strlen(source_filename)+6);
	if (!filename) exit (EXIT_FAILURE);
	strcpy (filename, source_filename);

	char* extension= find_ext(filename);

	/*------------------------------------------------------*\
	|*	Oeffne die Quelldatei				*|
	\*------------------------------------------------------*/

	if (!include (source_filename))
		exit (EXIT_FAILURE);

	/*------------------------------------------------------*\
	|*	Aufraeumen beim Programmende			*|
	\*------------------------------------------------------*/

	atexit (cleanup);

	/*------------------------------------------------------*\
	|*	Oeffne die Listingdatei				*|
	\*------------------------------------------------------*/

	if (getOption('l')) {

		if (!lstfilename) {
			strcpy (extension, ".lst");
			lstfilename = filename;
		}

		if (strcmp (lstfilename, "-"))
			lstfile = fopen (lstfilename, "w");
		else
			lstfile = stdout;
		if (!lstfile) {
			perror (lstfilename);
			exit (EXIT_FAILURE);
		}
	}

	/*------------------------------------------------------*\
	|*	Oeffne die Hexdatei				*|
	\*------------------------------------------------------*/

	if (!hexfilename) {
		strcpy (extension, ".hex");
		hexfilename = filename;
	}

	if (strcmp (hexfilename, "-"))
		hexfile = fopen (hexfilename, "w");
	else
		hexfile = stdout;
	if (!hexfile) {
		perror (hexfilename);
		exit (EXIT_FAILURE);
	}

	/*------------------------------------------------------*\
	|*	Fehlermeldungen ins listfile?			*|
	\*------------------------------------------------------*/

	if (chkOption('l','m')) errfile = lstfile;

	/*------------------------------------------------------*\
	|*	Uebersetze ein Programm				*|
	\*------------------------------------------------------*/

	DoProgram();

	/*------------------------------------------------------*\
	|*	Hexdatei: flush und Ende-Record			*|
	\*------------------------------------------------------*/

	CloseOut(0);

	/*------------------------------------------------------*\
	|*	Dateien schliessen				*|
	\*------------------------------------------------------*/

	errfile = stderr;

	if (lstfile && lstfile != stdout) fclose (lstfile);
	if (hexfile && hexfile != stdout) fclose (hexfile);

	/*------------------------------------------------------*\
	|*	Exit gut, alles gut.				*|
	\*------------------------------------------------------*/

	exit(EXIT_SUCCESS);

}


/*===================================================================*\
|*	$Id: nano.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
\*===================================================================*/
